# UpdateManager.py
# -*- Mode: Python; indent-tabs-mode: nil; tab-width: 4; coding: utf-8 -*-

import os
from apt import Cache
import sys
import shutil
import dbus
import logging
import dbus.service
from apt.debfile import DebPackage
from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)

from .Core.MyCache import MyCache
from .UpdateManagerDbus import UpdateManagerDbusController,UPDATER_DBUS_INTERFACE,UPDATER_DBUS_PATH,UPDATER_DBUS_SERVICE
from .Core.UpdateList import UpdateList
from .backend import InstallBackend,get_backend
from .Core.Database import Sqlite3Server
from .Core.loop import mainloop

import time
from gettext import gettext as _
from SystemUpdater.backend import DownloadBackend as downb
import apt_pkg
from SystemUpdater.Core.UpdaterConfigParser import UpgradeConfig
from SystemUpdater.Core.utils import get_broken_details,get_lis_from_cache

import subprocess

class UpdateManager():
    SELF_PKG_NAME = 'kylin-system-updater'
    GROUPS_PKG_NAME = 'kylin-update-desktop-config'
    INSTALL_ALONE_PROGRESS = "alone"

    def __init__(self,options):
        self.options = options
        self.cache = None
        self.update_list = None

        #表示后端的状态 注意：这个状态很重要 用不好整个后端代码就会卡住 
        self.is_working = InstallBackend.ACTION_DEFUALT_STATUS

        #建立dbus
        self.dbusController = self._setup_dbus()

        #连接数据库
        self.sqlite3_server = Sqlite3Server(self)

        #是否重启apt
        self.is_restart_aptdeamon = False

        #配置文件
        self.configs = UpgradeConfig("/var/lib/kylin-system-updater/")

    #检查是否需要重新启动aptdeamon 目前需要重启的有限速功能
    def restart_aptdeamon(self):
        if self.is_restart_aptdeamon == True:
            logging.warning("aptdeamon will be restart...")
            self.is_restart_aptdeamon = False
            self.dbusController.make_aptdeamon_restart()

    def run(self):
        """Start the daemon and listen for calls."""
        logging.debug("Waiting for calls")
        try:
            mainloop.run()
        except KeyboardInterrupt:
            self.dbusController.Quit(None)

    #进行清空所有下载的文件
    def start_clean(self):
        try:
            clean_backend = get_backend(self, InstallBackend.ACTION_CLEAN)
            clean_backend.start()
        except Exception as e:
            logging.error(e)

    #进行升级的操作
    def start_fix_broken(self):
        try:
            fix_backend = get_backend(self, InstallBackend.ACTION_FIX_BROKEN)
            fix_backend.start()
        except Exception as e:
            logging.error(e)
    
    #进行升级的操作
    def start_fix_incomplete(self):
        try:
            fix_backend = get_backend(self, InstallBackend.ACTION_FIX_INCOMPLETE)
            fix_backend.start()
        except Exception as e:
            logging.error(e)

    #download only
    def download_backend(self, pkgs):
        try:
            download_backend = downb.DownloadBackend(self)
            download_backend.downloadpkgs(pkgs)
        except Exception as e:
            logging.error(str(e))

    #download only 2.0
    def start_download(self, pkgs):
        try:
            if self.options.no_update_source is False:
                self.dbusController.on_update_important_list()

            update_backend = get_backend(self, InstallBackend.ACTION_DOWNLOADONLY)
            update_backend.start(partial_upgrade_list = pkgs)
        except Exception as e:
            logging.error(str(e))

    #进行更新的操作
    def start_update(self):
        _success = True
        header = ''
        desc = ''

        try:
            #进行检查网络
            self.dbusController.UpdateDetectStatusChanged(5,_("Checking network connection"))
            _success,header,desc = self.dbusController.check_connectivity()
            if _success == False:
                self.dbusController.DistUpdateDetectFinished(_success,[],header,desc)
                return

            #从服务器端更新摸板
            if self.options.no_update_source is False:
                self.dbusController.UpdateDetectStatusChanged(10,_("Updating Source Template"))
                _success,header,desc = self.dbusController.on_update_important_list()
                if _success == False:
                    self.dbusController.DistUpdateDetectFinished(_success,[],header,desc)
                    return
            #调用aptdeamon进行update
            update_backend = get_backend(self, InstallBackend.ACTION_UPDATE)
            update_backend.start()
        except Exception as e:
            logging.error(e)

    # 进行本地deb包安装的操作
    def start_deb_install(self, deb_path = "", _check_local_dep = False, _auto_satisfy = False):
        # _check_local_dep : 是否查询本地依赖
        # _auto_satisfy    : 是否通过网络下载依赖
        header = ''
        desc = ''
        absolute_path, debname = os.path.split(deb_path)
        try:
            # 是否有破损的包
            deb_cache = Cache()
            broken_count = deb_cache._depcache.broken_count
            if broken_count > 0:
            # 走 dpkg 安装流程，说明本地apt环境已经损坏
                dep_satisfy, header, desc = self._deb_install(deb_cache, deb_path, _check_local_dep)
                if dep_satisfy:
                    self.dbusController.InstalldebFinished(True, header, desc)
                else:
                    self.dbusController.InstalldebFinished(False, header, desc)
            else:
            # apt 安装流程
                dep_satisfy, header, desc = self._attempt_depends(deb_cache, deb_path, _check_local_dep,_auto_satisfy)
                if dep_satisfy:
                    try:
                        install_backend = get_backend(self, InstallBackend.ACTION_INSTALL_DEB)
                        install_backend.start_alone(partial_upgrade_list = deb_path, _is_install = _auto_satisfy)
                    except Exception as e:
                        logging.error(str(e))
                else:
                    self.dbusController.InstalldebFinished(False, header, desc)
        except Exception as e:
            logging.info(str(e))

    #进行升级的操作
    def start_install(self,upgrade_mode,is_install = False,partial_upgrade_list = []):
        try:
            if is_install == True:
                install_backend = get_backend(self, InstallBackend.ACTION_INSTALL)
                install_backend.start(upgrade_mode,partial_upgrade_list)
            else:
                resolver_backend = get_backend(self, InstallBackend.ACTION_CHECK_RESOLVER)
                resolver_backend.start(upgrade_mode,partial_upgrade_list)
        except Exception as e:
            logging.error(e)

    #进行删除的操作
    def start_purge_pkgs(self,pkgs_list):
        try:
            # 是否有破损的包
            deb_cache = Cache()
            broken_count = deb_cache._depcache.broken_count
            if broken_count > 0:
                _success,header,desc = self._dpkg_purge_pkgs(pkgs_list)
                if _success == True:
                    logging.info(header)
                    self.dbusController.PurgePackagesFinished(_success,'',desc)
                else:
                    self.dbusController.PurgePackagesFinished(_success,header,desc)
            else:
                purge_backend = get_backend(self, InstallBackend.ACTION_REMOVE_PACKAGES)
                purge_backend.start(partial_upgrade_list = pkgs_list)
        except Exception as e:
            logging.error(e)

    def _dpkg_purge_pkgs(self,pkgs_list):
        success = False
        pkg_list_str = " ".join(pkgs_list)
        args = ["dpkg", "-r"]
        args.extend([pkg_list_str])

        p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
        success = p.returncode == 0
        return success,p.stdout,''

    #更新结束之后会调到此获取要升级的列表 and 更新cache 生成升级组列表JSON
    def start_available(self):
        _success,header,desc = self.refresh_cache()

        #特殊情况的处理 单独安装包需要直接退出 安装or卸载执行完毕后 还会调到start_available
        if _success == False and header == self.INSTALL_ALONE_PROGRESS:
            return
        else:
            upgrade_list = []
            if self.update_list != None:
                upgrade_list = self.update_list.local_upgrade_data.upgrade_groups + self.update_list.local_upgrade_data.single_pkgs
            #发送更新升级列表完成的标志
            self.dbusController.DistUpdateDetectFinished(_success,upgrade_list,header,desc)
        #检查安装完成之后需要重启吗
        self.is_reboot_required()
        #清空cache
        if self.cache != None and self.cache.get_changes():
            self.cache.clear()

    def refresh_cache(self):
        _success = True
        header = ''
        desc = ''
        try:
            #第一次进入 之后update不进入
            if self.cache is None:
                self.cache = MyCache(None)
            else:
                self.cache.open(None)
                self.cache._initDepCache()
        except AssertionError:
            header = _("Software index is broken")
            desc = _("It is impossible to install or remove any software. "
                    "Please use the package manager \"Synaptic\" or run "
                    "\"sudo apt-get install -f\" in a terminal to fix "
                    "this issue at first.")
            _success = False
            return _success,header,desc

        except SystemError as e:
            header = _("Could not initialize the package information")
            desc = _("An unresolvable problem occurred while "
                    "initializing the package information.\n\n"
                    "Please report this bug against the 'kylin-system-updater' "
                    "package and include the following error "
                    "message:\n") + str(e)
            _success = False
            return _success,header,desc

        #提高生成对象的位置
        self.update_list = UpdateList(self)

        #1、 检查出现安装过程异常重启 出现的话 进行异常修复
        if self.configs.getWithDefault("SystemStatus", "isabnormalreboot", False) == True:
            header = self.INSTALL_ALONE_PROGRESS
            self.configs.setValue("SystemStatus","isabnormalreboot",str(False),True)
            logging.warning("start fix Abnormal Reboot broken pkgs...")
            self.start_fix_broken()
            return False,header,desc
        
        #检查当前系统的状态 是否存在系统为破损状态 导致出现异常
        self._check_cache_broken(self.cache)

        #检查自己是否需要升级更新
        _success,header,desc = self._check_self_upgrade(self.cache)
        if _success == False:
            return _success,header,desc

        #检查更新分组配置表
        _success,header,desc = self._check_group_config(self.cache)
        if _success == False and self.configs.getWithDefault("ConfigPkgStatus", "mustexistconfigpkg", True) == True:
            return _success,header,desc

        try:    
            _success,header,desc = self.update_list.update(self.cache)
        except SystemError as e:
            pass
        return _success,header,desc

    def _check_self_upgrade(self,cache):
        _success = True
        header = ''
        desc = ''
        #2、 安装JSON分组配置文件包 安装完毕会重新调start_available --> here 安装失败就直接退出不会进行下面的操作 
        try:
            if self.SELF_PKG_NAME in cache:
                pkg_json = cache[self.SELF_PKG_NAME]
                #是否安装
                if pkg_json.is_installed and pkg_json.is_upgradable:
                    logging.info("Self-Updater(%s) start upgrading...",self.SELF_PKG_NAME)
                    pkg_json.mark_upgrade()

                    self.start_install(InstallBackend.MODE_UPGRADE_SINGLE,True)
                    #直接退出 进行安装程序
                    _success = False
                    header = self.INSTALL_ALONE_PROGRESS
                    return _success,header,desc
                else:
                    logging.info("Self-Updater(%s) No need to upgrade...",self.SELF_PKG_NAME)
                    #直接退出 进行之后的动作
                    _success = True
                    return _success,header,desc
            else:
                #没有在cache中就认为不需要升级
                logging.error("Self-Updater(%s) The upgrade package is not in Cache...",self.SELF_PKG_NAME)
                #直接退出 进行之后的动作
                _success = True
                return _success,header,desc                

        #FIXME: 错误处理未做 报告到控制面板
        except Exception as e:
            logging.error(e)
            _success = True
            return _success,header,desc

    def _check_group_config(self,cache):
        _success = True
        header = ''
        desc = ''
        #2、 安装JSON分组配置文件包 安装完毕会重新调start_available --> here 安装失败就直接退出不会进行下面的操作 
        try:
            if self.GROUPS_PKG_NAME in cache:
                pkg_json = cache[self.GROUPS_PKG_NAME]
                #是否安装
                if pkg_json.is_installed:
                    #是否可升级
                    if pkg_json.is_upgradable:
                        logging.info("groups JSON ConfigPkgs(%s) start upgrading...",self.GROUPS_PKG_NAME)
                        pkg_json.mark_upgrade()

                        self.start_install(InstallBackend.MODE_UPGRADE_SINGLE,True)
                        #直接退出
                        _success = False
                        header = self.INSTALL_ALONE_PROGRESS
                        return _success,header,desc
                    else:
                        logging.info("ConfigPkgs(%s) No need to upgrade...",self.GROUPS_PKG_NAME)
                else:
                    logging.info("groups JSON ConfigPkgs(%s) start new installing...",self.GROUPS_PKG_NAME)
                    pkg_json.mark_install()
                    self.start_install(InstallBackend.MODE_UPGRADE_SINGLE,True)
                    #直接退出
                    _success = False
                    header = self.INSTALL_ALONE_PROGRESS
                    return _success,header,desc
            else:
                header = _("Unable to get group configuration package, please contact administrator to solve")
                desc = _("groups JSON ConfigPkgs install failed")
                logging.error("groups JSON ConfigPkgs(%s) install failed...",self.GROUPS_PKG_NAME)
                _success = False
                return _success,header,desc                

        #FIXME: 错误处理未做 报告到控制面板
        except Exception as e:
            logging.error(e)
            _success = False
            return _success,header,desc
        
        #FIXME:
        #3、 判断目录是JSON配置文件夹是否缺失 缺失后进行修复 卸载重新安装步骤
        # if not os.path.exists(INPUT_CONFIG_PATH):
        #     logging.info("groups JSON Config Path(%s) Missing and Trying to fix...",INPUT_CONFIG_PATH)
        #     #将软件包卸载 之后进行重新安装here --> purge --> start_available 进行判断是否安装未安装重新安装
        #     # self.start_install_alone(pkgs_purge = [self.GROUPS_PKG_NAME])
        #     #直接退出
        #     _success = False
        #     header = self.INSTALL_ALONE_PROGRESS
        #     return _success,header,desc

        return _success,header,desc

    def is_reboot_required(self):
        """If a reboot is required to get all changes into effect."""
        return os.path.exists(os.path.join(apt_pkg.config.find_dir("Dir"),
                                        "var/run/reboot-required"))

    def _check_cache_broken(self,cache):
        with cache.actiongroup():
            if cache.get_changes():
                cache.clear()
            #获取出现破损状态包的数量
            wouldDelete = cache._depcache.broken_count
            
            if wouldDelete > 0:
                #获取那些依赖关系不满足导致的问题
                logging.info(get_broken_details(cache))
            else:
                return
            
            try:
                #计算依赖的解决方案
                cache.fix_broken()
            except SystemError as e:
                logging.error(e)
                return
            pkgs_install,pkgs_upgrade,pkgs_remove = get_lis_from_cache(cache)
            logging.warning("Fix broken Packages is need to install:%a and upgrade:%a and remove:%a",pkgs_install,pkgs_upgrade,pkgs_remove)

            if cache.get_changes():
                cache.clear()

    def _setup_dbus(self):
        # check if there is another g-a-i already and if not setup one
        # listening on dbus
        try:
            bus = dbus.SystemBus()
        except Exception:
            logging.error("warning: could not initiate dbus")
            return
        try:
            proxy_obj = bus.get_object(UPDATER_DBUS_SERVICE,
                                        UPDATER_DBUS_PATH)

            #替换后台
            if self.options.replace is False:
                logging.critical("Another daemon is already running")
                sys.exit(1)
            else:
                logging.warning("Replacing already running daemon")
                proxy_obj.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                    timeout=300)
                time.sleep(1)
                bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                                bus,
                                                do_not_queue=True)
                return  UpdateManagerDbusController(self, bus_name)

        except dbus.DBusException:
            bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                            bus)
            logging.info(_("initiate dbus success ..."))
            return  UpdateManagerDbusController(self, bus_name)

    # 是否查找本地依赖
    def _attempt_depends(self,deb_cache, deb_path,_check_local_dep,_auto_satisfy):  
        depends_list = []
        depends_pkg = []
        satisfy_list = []
        depends_count = 0
        _local_satisfy = False
        error_string = ''
        error_desc = ''
        absolute_path, debname = os.path.split(deb_path)
        try:
            deb = DebPackage(deb_path, deb_cache)
            deb.check()
            (install, remove, unauth) = deb.required_changes
        except Exception as e:
            logging.error(str(e))
            return False,str(e), error_desc
        # 依赖不满足的情况
        if len(install) > 0:
            if _check_local_dep: #查找本地
                if remove:
                    logging.error("Need uninstall: %s.",str(remove))
                    error_string = "Installing "+str(debname.split("_")[0])+" requires uninstalling: "+str(remove)
                    logging.error(error_string)
                # 需要查找本地依赖
                elif len(install) > 0:
                    for pkg in deb_cache:
                        if pkg.marked_install and pkg.name != str(debname.split("_")[0]):
                            depends_pkg.append(pkg)
                        elif pkg.marked_upgrade and pkg.name != str(debname.split("_")[0]):
                            depends_pkg.append(pkg)
                if len(depends_pkg)>0: #查找本地deb包
                    depends_list = [debfile for debfile in os.listdir(absolute_path) if debfile.endswith(".deb")]
                    for depends in depends_pkg:
                        for debfile in depends_list:
                            if depends.name in debfile and depends.candidate.version in debfile:
                                #FIXME:检查depends包的合法性
                                depends_count += 1
                                satisfy_list.append(debfile)
                    if depends_count < len(depends_pkg):
                        #本地依赖不满足 判断源下载
                        if _auto_satisfy:
                            _local_satisfy = True
                        elif not _auto_satisfy:
                            _local_satisfy = False
                            error_string = str(debname.split("_")[0])+" dependency is not satisfied. "
                            error_desc   = ",".join(install)
                            logging.error(error_string+ error_desc)
                    else:
                        #将应用包与依赖包拷贝至archive目录安装
                        try:
                            if debname not in satisfy_list:
                                satisfy_list.append(debname)
                            for satisfy in satisfy_list:
                                shutil.copy(os.path.join(absolute_path,satisfy),"/var/cache/apt/archives/")
                                logging.info("move debpkg: %s",satisfy)
                            _local_satisfy = True
                        except Exception as e:
                            logging.info(str(e))
                return _local_satisfy,error_string,error_desc
            elif not _check_local_dep and _auto_satisfy:
                _local_satisfy = True
                if install:
                    error_string = str(debname.split("_")[0])+" dependency is not satisfied, will download. "
                    error_desc   = ",".join(install)
                    logging.error(error_string+error_desc)
                    return _local_satisfy,error_string,error_desc
            elif not _check_local_dep and not _auto_satisfy:
                _local_satisfy = False
                if install:
                    error_string = str(debname.split("_")[0])+" dependency is not satisfied. "
                    error_desc   = ",".join(install)
                    logging.error(error_string+error_desc)
                    return _local_satisfy,error_string,error_desc
        # 依赖满足
        else:
            _local_satisfy = True
            error_string = ''
            error_desc=''
            return _local_satisfy,error_string,error_desc
    
    def _deb_install(self, deb_cache, deb_path, _check_local_dep):
        depends_list = []
        depends_pkg = []
        satisfy_list = []
        depends_count = 0
        _status = False
        error_string = ''
        header = ''
        desc = ''
        absolute_path, debname = os.path.split(deb_path)
        try:
            deb = DebPackage(deb_path, deb_cache)
            deb.check()
            (install, remove, unauth) = deb.required_changes
        except Exception as e:
            logging.error(str(e))
            return False,str(e), desc
         # 依赖不满足的情况
        if len(install) > 0:
            if _check_local_dep: #查找本地
                # 需要查找本地依赖
                if len(install) > 0:
                    for pkg in deb_cache:
                        if pkg.marked_install and pkg.name != str(debname.split("_")[0]):
                            depends_pkg.append(pkg)
                        elif pkg.marked_upgrade and pkg.name != str(debname.split("_")[0]):
                            depends_pkg.append(pkg)
                if len(depends_pkg)>0: #查找本地deb包
                    depends_list = [debfile for debfile in os.listdir(absolute_path) if debfile.endswith(".deb")]
                    for depends in depends_pkg:
                        for debfile in depends_list:
                            if depends.name in debfile and depends.candidate.version in debfile:
                                depends_count += 1
                                satisfy_list.append(debfile)
                    if depends_count < len(depends_pkg):
                        #本地依赖不满足
                        _status = False
                        error_string = str(debname.split("_")[0])+" dependency is not satisfied. "+",".join(install)
                        logging.error(error_string)
                        header = error_string
                        return False,header, desc
                    else:
                        #安装依赖包
                        if debname not in satisfy_list:
                            satisfy_list.append(debname)
                        for satisfy in satisfy_list:
                            try:
                                deb = DebPackage(os.path.join(absolute_path,satisfy))
                                ret = deb.install()
                            except Exception as e:
                                logging.error(e)
                                return False, str(e), desc
                        return True, header, desc
            elif not _check_local_dep:
                _status = False
                if install:
                    error_string = str(debname.split("_")[0])+" dependency is not satisfied. "+",".join(install)
                    logging.error(error_string)
                    header = error_string
                    return False,header, desc
        # 依赖满足
        else:
            try:
                deb = DebPackage(deb_path)
                ret = deb.install()
            except Exception as e:
                logging.error(e)
                return False, str(e), desc
            return True, header, desc